#include <time.h>
#ifdef _OPENMP
#include <omp.h>
#endif
#include "run.h"

int map_unordered = 0;

#define USE_EXTERNAL_HIT_POINTER

#include "stats.h"


#ifdef __AVX2__

#define MAX_MULTIPLE_HITS 8

const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {8, 8, 8, 8, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, int *, unsigned char *) = {
  &alignment_avx2__align_octa,
  &alignment_avx2__align_octa,
  &alignment_avx2__align_octa,
  &alignment_avx2__align_octa,
  &alignment_avx2__align_quad,
  &alignment_avx2__align_quad,
  &alignment_avx2__align_quad,
  &alignment_avx2__align_quad,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair,
  &alignment_avx2__align_pair
};
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
  &alignment_avx2__init_octa,
  &alignment_avx2__init_octa,
  &alignment_avx2__init_octa,
  &alignment_avx2__init_octa,
  &alignment_avx2__init_quad,
  &alignment_avx2__init_quad,
  &alignment_avx2__init_quad,
  &alignment_avx2__init_quad,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair,
  &alignment_avx2__init_pair
};
const int simd_N_BYTE_table[INDEL_DATA_VECTOR_SIZE] = {4, 4, 4, 4, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};
#else
#ifdef __SSE2__

#define MAX_MULTIPLE_HITS 4

const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {4, 4, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, int *, unsigned char *) = {
  &alignment_sse2__align_quad,
  &alignment_sse2__align_quad,
  &alignment_sse2__align_quad,
  &alignment_sse2__align_quad,
  &alignment_sse2__align_pair,
  &alignment_sse2__align_pair,
  &alignment_sse2__align_pair,
  &alignment_sse2__align_pair,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono,
  &alignment_sse2__align_mono
};
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
  &alignment_sse2__init_quad,
  &alignment_sse2__init_quad,
  &alignment_sse2__init_quad,
  &alignment_sse2__init_quad,
  &alignment_sse2__init_pair,
  &alignment_sse2__init_pair,
  &alignment_sse2__init_pair,
  &alignment_sse2__init_pair,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono,
  &alignment_sse2__init_mono
};
const int simd_N_BYTE_table[INDEL_DATA_VECTOR_SIZE] = {4, 4, 4, 4, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};
#else
#ifdef __SSE__

#define MAX_MULTIPLE_HITS 4

const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {4, 4, 2, 2, 1, 1, 1, 1};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, int *, unsigned char *) = {
  &alignment_sse__align_quad,
  &alignment_sse__align_quad,
  &alignment_sse__align_pair,
  &alignment_sse__align_pair,
  &alignment_sse__align_mono,
  &alignment_sse__align_mono,
  &alignment_sse__align_mono,
  &alignment_sse__align_mono
};
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
  &alignment_sse__init_quad,
  &alignment_sse__init_quad,
  &alignment_sse__init_pair,
  &alignment_sse__init_pair,
  &alignment_sse__init_mono,
  &alignment_sse__init_mono,
  &alignment_sse__init_mono,
  &alignment_sse__init_mono
};
const int simd_N_BYTE_table[INDEL_DATA_VECTOR_SIZE] = {2, 2, 4, 4, 8, 8, 8, 8};
#else

#define MAX_MULTIPLE_HITS 1

void fake_init (unsigned int match,
                unsigned int mismatch,
                unsigned int gapopen,
                unsigned int gapextends,
                unsigned int threshold,
                unsigned int readlength) {}
int fake_align (unsigned char * genome,
                int * pos_genome,
                unsigned char * read) { return 1; /* 0x01 */}

const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {1, 1, 1, 1, 1, 1, 1, 1};
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
  &fake_init,
  &fake_init,
  &fake_init,
  &fake_init,
  &fake_init,
  &fake_init,
  &fake_init,
  &fake_init
};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, int *, unsigned char *) = {
  &fake_align,
  &fake_align,
  &fake_align,
  &fake_align,
  &fake_align,
  &fake_align,
  &fake_align,
  &fake_align
};
const int simd_N_BYTE_table[INDEL_DATA_VECTOR_SIZE] = {1, 1, 1, 1, 1, 1, 1, 1};
#endif
#endif
#endif

int SHOW_TRACEBACK_PATTERNS = 0;
int LIST_UNMAPPED_READS = 0;

/*
 * #define __DONT__MAP__
 */

/**
 * Statistics for finding the efficiency of the filters
 */

/*
 * #define ENABLE_STATS
 */

#ifdef ENABLE_STATS

#define IDX_PROCESSED_READS 0
#define IDX_SEED_HITS       1
#define IDX_SIMD_PASSED     2
#define IDX_ALGO_PASSED     3
#define STATS_FIELDS_COUNT  4


int FILTER_STATS[4][4] = {
  {0,0,0,0},
  {0,0,0,0},
  {0,0,0,0},
  {0,0,0,0}
};

int *stats__seed_idx;

void stats__init() {
#ifdef _OPENMP
  SAFE_FAILURE__ALLOC(stats__seed_idx, MAX_THREADS, int);
#else
  SAFE_FAILURE__ALLOC(stats__seed_idx, 1, int);
#endif
}

#ifdef _OPENMP
#define stats__set_seed_idx(seed_idx) {stats__seed_idx[_ogtn] = seed_idx;}
#else
#define stats__set_seed_idx(seed_idx) {stats__seed_idx[0] = seed_idx;}
#endif


#ifdef _OPENMP
#define stats__inc(field_idx,count)  FILTER_STATS[stats__seed_idx[_ogtn]][field_idx] += count;
#else
#define stats__inc(field_idx,count)  FILTER_STATS[stats__seed_idx[0]][field_idx] += count;
#endif


void stats__destroy() {
  free(stats__seed_idx);
}

void stats__display(IndexType* index, int* stats) {
#ifdef DEBUG_INDEX
  INFO__(("%s:\n  REAL HITS:      %10d",              seed__to_string(index->seed), index__get_hits_count(index)));
  INFO__((     "  EVALUATED HITS: %10d (%3.2lf%%)",   stats[IDX_SEED_HITS],   stats[IDX_SEED_HITS]   * 100.0 / index->hits_count     ));
#else
  INFO__((     "  EVALUATED HITS: %10d",   stats[IDX_SEED_HITS] ));
#endif
  INFO__((     "  SIMD FILTER:    %10d (%3.2lf%%)",   stats[IDX_SIMD_PASSED], stats[IDX_SIMD_PASSED] * 100.0 / stats[IDX_SEED_HITS]  ));
  INFO__((     "  FULL ALIGNMENT: %10d (%3.2lf%%)\n", stats[IDX_ALGO_PASSED], stats[IDX_ALGO_PASSED] * 100.0 / stats[IDX_SIMD_PASSED]));
}

#endif

/**
 * Perform an alignment between two sequences
 * @param read_str First sequence (color code)
 * @param ref_str Second sequence (color code)
 * @param ref_masked_str Second sequence mask
 * @param read quality values for the first sequence
 * @param match
 * @param mismatch
 * @param gap_open
 * @param gap_extend
 * @param allowed_indels
 * @param output The output file descriptor
 */
int single_alignment(const char* read_str, const char* ref_str, const char* ref_masked_str, const char* qual_str,
                     const ScoreType match, const ScoreType mismatch, const ScoreType gap_open, const ScoreType gap_extend, const int allowed_indels, const int simd_allowed_diags,
                     FILE* output) {
  CODE_TYPE* read        = string_to_code(read_str);
  CODE_TYPE* ref         = string_to_code(ref_str);
  CODE_TYPE* ref_masked  = string_to_code(ref_masked_str);
  int read_len = strlen(read_str);
  short *quality = NULL;
  ScoreType best;
  AlignmentType* alignment;

  VERBOSITY = VERBOSITY_MODERATE; // to allow the displayed alignment to be built

  if (qual_str) {
    quality = parse_integer_sequence(qual_str, read_len);
  }

  alignment = alignment__create_single(read, quality, strlen(read_str), ref, ref_masked, strlen(ref_str), allowed_indels);
  alignment__set_params(alignment, match, mismatch, gap_open, gap_extend);

  best = alignment__align(alignment, MIN_SCORE);

  fprintf(output, "\n%s\n", alignment__traceback(alignment));
  fprintf(output, "Score: %d\n\n", best);
  alignment__destroy(alignment);
  free(read);
  free(ref);
  if (quality) {
    free(quality);
  }

  return best;
}

/*------------------------------------------------------------------------------------------------------
 * For multithreaded execution via openmp, some data structures need to be cloned for each thread
 */
#ifdef _OPENMP

#define QUALITY_MASK__DECLARE        unsigned int ** quality_mask;

#define QUALITY_MASK__INIT(len) {                                                    \
    int t;                                                                           \
    SAFE_FAILURE__ALLOC(quality_mask, MAX_THREADS, unsigned int*);                   \
    for (t = 0; t < MAX_THREADS; ++t) {                                              \
      SAFE_FAILURE__ALLOC(quality_mask[t], COMPRESSED_QSEED_LEN(len), unsigned int); \
      memset(quality_mask[t], 0x00, COMPRESSED_QSEED_LEN(len)*sizeof(unsigned int)); \
    }                                                                                \
  }

#define QUALITY_MASK_REF   quality_mask[_ogtn]

#define QUALITY_MASK__DESTROY  {                \
    int t;                                      \
    for (t = 0; t < MAX_THREADS; ++t) {         \
      free(quality_mask[t]);                    \
    }                                           \
    free(quality_mask);                         \
  }


/* Alignment structures for each process */

#define ALIGNMENT__DECLARE(alignment) AlignmentType* alignment;

#define ALIGNMENT__INIT(alignment, read_len, ref_len, allowed_indels, match, mismatch, gap_open, gap_extend)  { \
    int t;                                                                                                      \
    SAFE_FAILURE__ALLOC(alignment, MAX_THREADS, AlignmentType);                                                 \
    for (t = 0; t < MAX_THREADS; ++t) {                                                                         \
      alignment__init(&alignment[t], read_len, ref_len, allowed_indels);                                        \
      alignment__set_params(&alignment[t], match, mismatch, gap_open, gap_extend);                              \
    }                                                                                                           \
  }

#define ALIGNMENT__RESET_READ(alignment, read, quality) {                    \
    alignment__reset_with_compressed_read(&alignment[_ogtn], read, quality); \
  }

#define ALIGNMENT__REF(alignment) &alignment[_ogtn]

#define ALIGNMENT__DESTROY(alignment) {   \
    int t;                                \
    for (t = 0; t < MAX_THREADS; ++t) {   \
      alignment__destroy(&alignment[t]);  \
    }                                     \
    free(alignment);                      \
  }


/* Filters for each process */

#define FILTER__DECLARE(filter) RegionFilterType* filter;

#define FILTER__INIT(filter, ref_db, ref_window) {              \
    int t;                                                      \
    SAFE_FAILURE__ALLOC(filter, MAX_THREADS, RegionFilterType); \
    for (t = 0; t < MAX_THREADS; ++t) {                         \
      reg_filter__init(&filter[t], ref_db, ref_window);         \
    }                                                           \
  }

#define FILTER__REF(filter) &filter[_ogtn]

#define FILTER__DESTROY(filter) {               \
    int t;                                      \
    for (t = 0; t < MAX_THREADS; ++t) {         \
      reg_filter__destroy(&filter[t]);          \
    }                                           \
    free(filter);                               \
  }

/* Reverse complementary data */
#define REV_SEQ__DECLARE(tmp_sequence)  ReadDataType *tmp_sequence;

#define REV_SEQ__ALLOC(tmp_sequence, len) {                             \
    int t;                                                              \
    SAFE_FAILURE__ALLOC(tmp_sequence, MAX_THREADS, ReadDataType);       \
    for (t = 0; t < MAX_THREADS; ++t){                                  \
      SAFE_FAILURE__ALLOC( tmp_sequence[t].sequence, len, CODE_TYPE);   \
      tmp_sequence[t].quality = NULL;                                   \
    }                                                                   \
  }

#define REV_SEQ__REF(tmp_sequence)     &tmp_sequence[_ogtn]

#define REV_SEQ__SEQ_REF(tmp_sequence)  tmp_sequence[_ogtn].sequence

#define REV_SEQ__QUAL_REF(tmp_sequence) tmp_sequence[_ogtn].quality

#define REV_SEQ__DESTROY(tmp_sequence) {        \
    int t;                                      \
    for (t = 0; t < MAX_THREADS; ++t){          \
      free(tmp_sequence[t].sequence);           \
      if (tmp_sequence[t].quality) {            \
        free(tmp_sequence[t].quality);          \
      }                                         \
    }                                           \
    free(tmp_sequence);                         \
  }

#else

#define QUALITY_MASK__DECLARE        unsigned int * quality_mask;

#define QUALITY_MASK__INIT(len) {                                               \
    SAFE_FAILURE__ALLOC(quality_mask, COMPRESSED_QSEED_LEN(len), unsigned int); \
    memset(quality_mask, 0x00, COMPRESSED_QSEED_LEN(len)*sizeof(unsigned int)); \
  }

#define QUALITY_MASK_REF             quality_mask

#define QUALITY_MASK__DESTROY        {free(quality_mask);}


/* Alignment Structure */

#define ALIGNMENT__DECLARE(alignment) AlignmentType alignment;

#define ALIGNMENT__INIT(alignment, read_len, ref_len, allowed_indels, match, mismatch, gap_open, gap_extend) { \
    alignment__init(&alignment, read_len, ref_len, allowed_indels);                                            \
    alignment__set_params(&alignment, match, mismatch, gap_open, gap_extend);                                  \
  }

#define ALIGNMENT__RESET_READ(alignment, read, quality) {               \
    alignment__reset_with_compressed_read(&alignment, read, quality);   \
  }

#define ALIGNMENT__REF(alignment) &alignment

#define ALIGNMENT__DESTROY(alignment) alignment__destroy(&alignment);

/* Filter */
#define FILTER__DECLARE(filter) RegionFilterType filter;

#define FILTER__INIT(filter, ref_db, ref_window) reg_filter__init(&filter, ref_db, ref_window);

#define FILTER__REF(filter) &filter

#define FILTER__DESTROY(filter) reg_filter__destroy(&filter);

/* Reverse complementary data */
#define REV_SEQ__DECLARE(tmp_sequence)  ReadDataType tmp_sequence;

#define REV_SEQ__ALLOC(tmp_sequence, len) SAFE_FAILURE__ALLOC(tmp_sequence.sequence, len, CODE_TYPE); tmp_sequence.quality = NULL;

#define REV_SEQ__REF(tmp_sequence)  &tmp_sequence

#define REV_SEQ__SEQ_REF(tmp_sequence)  tmp_sequence.sequence

#define REV_SEQ__QUAL_REF(tmp_sequence) tmp_sequence.quality

#define REV_SEQ__DESTROY(tmp_sequence) { free(tmp_sequence.sequence); if (tmp_sequence.quality) {free(tmp_sequence.quality);}}

#endif


/*---------------------------------------------->8--------------------------------------------------------*/

/**
 * Find and register all the hits of applying a certain seed at a given position in a read
 * @param read
 * @param read_id The read's index (for updating the map)
 * @param ref_id The reference's index (for updating the map)
 * @param j The position where the seed should be applied
 * @param filter Constrains references to the seed, the index AND the reference database
 * @param alignment The alignment data structure
 * @param map the read map to update with good alignments
 */

static inline void process_read_with_seed(
                                          CODE_TYPE* read_sequence, const int read_id, const int ref_id, const int j, const int simd_allowed_diags,
                                          RegionFilterType* filter, IndexType* index,
                                          AlignmentType* alignment, HitMapType* map,
                                          const int alignment_sense) {
  int updates = 0;
  int I = alignment->params.allowed_indels;


  CODE_TYPE* ref_seq         = ((ReferenceDBType*)index->db)->sequence;
  CODE_TYPE* ref_seq_masked  = ((ReferenceDBType*)index->db)->sequence_masked;
  ScoreType score, score_threshold = min_accepted_score;

  int hits[MAX_MULTIPLE_HITS] = {0};
  int ref_window_simd[MAX_MULTIPLE_HITS] = {0};

#if (!defined USE_EXTERNAL_HIT_POINTER) || (defined ENABLE_STATS)
#ifdef _OPENMP
  const int _ogtn = omp_get_thread_num();
#else
  const int _ogtn = 0;
#endif
#endif

  /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
  stats__inc(IDX_PROCESSED_READS, 1);
#endif
  // get the resulted code key
  int key = seed__apply_to_compressed(index->seed, read_sequence, j);

#ifdef USE_EXTERNAL_HIT_POINTER
  int hit_pointer = -1;
#else
  index__reset_crt_hit(index, _ogtn);
#endif

  /* search it in the index i */
  int h = 0;
  while (h != -1) {
#ifdef USE_EXTERNAL_HIT_POINTER
    h = reg_filter__extern_next_hits(filter, index, key, hits, simd_mul[simd_allowed_diags], &hit_pointer);
#else
    h = reg_filter__next_hits(filter, index, key, hits, simd_mul[simd_allowed_diags], _ogtn);
#endif
    /* no hit at all? exit loop */
    if (hits[0] == -1) {
      break;
    }
    /* simd filter */
    int i;
    for (i = 0; i < simd_mul[simd_allowed_diags]; ++i) {
      if (i == 0 || hits[i] != hits[i-1]) {
        ref_window_simd[i] = hits[i] - j - simd_allowed_diags;

        /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
        stats__inc(IDX_SEED_HITS, 1);
#endif
        VERB_FILTER(VERBOSITY_ANNOYING, INFO__(("\nRead %d[%d]: HIT Reference@%d", read_id, j, hits[i])););
      } else {
        /* copy to do "same work" on unused diagonal block */
        ref_window_simd[i] = ref_window_simd[i-1];
      }
    }

    /* fast filtering using the SIMD alignment functions */
    int x = simd_fct_table[simd_allowed_diags](ref_seq, ref_window_simd, read_sequence);
    int p = 0;

    /* check which alignment has a hit (if any, thus when x != 0)*/
    while (x != 0) {
      if((x & 1) && (p == 0 || hits[p] != hits[p-1])) {

        int ref_window = max(hits[p] - j - I, -simd_allowed_diags);

        VERB_FILTER(VERBOSITY_ANNOYING, INFO__(("\nRead %d PASSED SIMD FILTER Reference@%d", read_id, ref_window)););

        /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
        stats__inc(IDX_SIMD_PASSED, 1);
#endif
        alignment__reset_with_compressed_ref(alignment, ref_seq, ref_seq_masked, ref_window);
        //If the hit is not in a region that has already been checked, and the alignment score is acceptable, memorize it
        if ((score = alignment__align(alignment, score_threshold)) > score_threshold) {

          /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
          stats__inc(IDX_ALGO_PASSED, 1);
#endif
          alignment__traceback(alignment);
          // mask the aligned zone
          reg_filter__mask(filter, alignment, ref_window);

          // display the alignment
          VERB_FILTER(VERBOSITY_HIGH, INFO__(("\nReference@%d, Read %d\n%sScore: %d\n", ref_window + alignment->best_j0, read_id, alignment->to_display, score)););
          // memorize in a map
#ifdef _OPENMP
#pragma omp critical(mask)
#endif
          {
            updates += (hit_map__update(map, read_id, ref_id, ref_window, alignment, alignment_sense) >= 0) ? 1 : 0;
          }
          if (map->map[read_id][MAP_DETAIL_SIZE-1].score > score_threshold) {
            score_threshold = map->map[read_id][MAP_DETAIL_SIZE-1].score;
          }
        }
      }
      /* next alignment in the SIMD result (binary flag output)*/
      x >>= 1;
      ++p;
    }
  }
}

/**
 * Reads mapped on a reference
 * @param reads_filename
 * @param qual_filename
 * @param ref_filename
 * @param seed_list
 * @param match
 * @param mismatch
 * @param gap_open
 * @param gap_extend
 * @param allowd_indels
 * @param output
 */
int reads_against_reference(const char* reads_filename, const char* qual_filename, const char* ref_filename,
                            const char* seedslist,
                            const ScoreType match, const ScoreType mismatch, const ScoreType gap_open, const ScoreType gap_extend, const int allowed_indels, const int simd_allowed_diags,
                            FILE* output) {
  ReadsDBType       reads_db;
  ReferenceDBType*  ref_dbs;
  int               ref_dbs_size;
  IndexType**       ref_index;
  SeedType **       seeds;


#ifdef _OPENMP
  omp_set_dynamic(0);
  omp_set_num_threads(MAX_THREADS);
#endif

  ALIGNMENT__DECLARE(alignment);
  FILTER__DECLARE(filter);

  REV_SEQ__DECLARE(tmp_sequence);

  QUALITY_MASK__DECLARE;

  HitMapType* map;

  time_t start_time = time(NULL), crt_time, crt_time1;

#ifdef ENABLE_STATS
  stats__init();
#endif

  /* global variable used to allocate the correct number of byte after each ALLOC, in order to enable block reading without alarming valgrind */
  N_BYTES = simd_N_BYTE_table[simd_allowed_diags];

  /* create seeds */
  int seed_count = seed__parse_list(seedslist, &seeds);
  /* verify seed validity */
  if (seed_count == RETURN_INPUT_ERR) {
    ERROR__(("The seed pattern is invalid: \"%s\".\nExpected: {0,1}+ or {-,#}+ with at least one '1'/'#' (seed separator ';').", seedslist));
    exit (RETURN_INPUT_ERR);
  } else if (seed_count <= 0) {
    ERROR__(("The seed input is invalid: \"%s\".\nPlease provide at least one seed.", seedslist));
    exit (RETURN_INPUT_ERR);
  }
  VERB_FILTER(VERBOSITY_MODERATE,
            INFO__(("The following %d seed%s will be used:\n", seed_count, (seed_count > 1 ? "s" : "")));
            int si; for (si = 0; si < seed_count; ++si) seed__display(seeds[si]);
            printf("\n");
            );

  /* load the two databases */
  /* load the reads */
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loading reads database...\n"););
  crt_time = time(NULL);
  if (load_reads_db(reads_filename, qual_filename, &reads_db) <= 0) {
    ERROR__(("Reads could not be loaded. Exiting."));
    exit (RETURN_INPUT_ERR);
  }
  map = hit_map__create(reads_db.size, allowed_indels);
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loaded %d reads (read length %d) in %ld seconds.\n\n", reads_db.size, reads_db.read_len, time(NULL) - crt_time););

  /* sort reads in alphanumeric order */
  VERB_FILTER(VERBOSITY_MODERATE, printf("Sorting reads database...\n"););
  crt_time = time(NULL);
  sort_reads_db(&reads_db);
  VERB_FILTER(VERBOSITY_MODERATE, printf("Sorted %d reads in %ld seconds.\n\n", reads_db.size, time(NULL) - crt_time););

  /* load the reference */
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loading reference database...\n"););
  crt_time = time(NULL);
  if ((ref_dbs_size = load_reference_db(ref_filename, reads_db.read_len, &ref_dbs)) <= 0) {
    ERROR__(("Reference could not be loaded. Exiting."));
    exit(RETURN_INPUT_ERR);
  }
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loaded %d reference sequence%s in %ld seconds.\n\n", ref_dbs_size, (ref_dbs_size == 1 ? "" : "s"), time(NULL) - crt_time););

  /* create alignment data */
  int ref_window_len = reads_db.read_len + (allowed_indels<<1);
  /* SIMD filter initialization */
  simd_init_fct_table[simd_allowed_diags](abs(match),abs(mismatch),abs(gap_open),abs(gap_extend),min_accepted_score_simd_filter, reads_db.read_len);

  ALIGNMENT__INIT(alignment, reads_db.read_len, ref_window_len, allowed_indels, match, mismatch, gap_open, gap_extend);
  REV_SEQ__ALLOC(tmp_sequence, reads_db.read_len);
  /* min seed length, for masking... */
  int min_seed_len = reads_db.read_len;


  {
    int rfi;
    for (rfi = 0; rfi < seed_count; ++rfi) {
      if (min_seed_len > seeds[rfi]->length) {
        min_seed_len = seeds[rfi]->length;
      }
    }
  }

  /* Create indexes */
  SAFE_FAILURE__ALLOC(ref_index, seed_count, IndexType*);
  /* Quality mask for "seed on read" masking */
  QUALITY_MASK__INIT(reads_db.read_len);

  crt_time = time(NULL);

  /* Process each reference sequence */
  int ref_id;
  for (ref_id = 0; ref_id < ref_dbs_size; ++ref_id) {
    ReferenceDBType* ref_db = &ref_dbs[ref_id];
#ifdef NUCLEOTIDES
    VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\n\nProcessing reference sequence %s (%d of %d, %d bases)...\n", ref_db->name?ref_db->name:"(null)", ref_id + 1, ref_dbs_size, ref_db->size)););
#else
    VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\n\nProcessing reference sequence %s (%d of %d, %d colors)...\n", ref_db->name?ref_db->name:"(null)", ref_id + 1, ref_dbs_size, ref_db->size)););
#endif

    /* create index */
    VERB_FILTER(VERBOSITY_MODERATE, printf("Creating reference index...\n"););
    crt_time1 = time(NULL);
    int si;
#ifdef _OPENMP
#pragma omp parallel for shared(ref_index, ref_window_len, ref_db, seeds)
#endif
    for (si = 0; si < seed_count; ++si) {
      ref_index[si] = index__build_reference(ref_db, seeds[si]);
    }
    VERB_FILTER(VERBOSITY_MODERATE, printf("Created %d indexes in %ld seconds.\n\n", seed_count, time(NULL) - crt_time1););

    /* create filter */
    FILTER__INIT(filter, ref_db, (ref_window_len - min_seed_len));

    /* On each read, pas each seed, find the hits, and align */
    VERB_FILTER(VERBOSITY_MODERATE, printf("Start search...\n"););

    crt_time1 = time(NULL);
    if (ALIGNMENT_SENSE & ALIGNMENT_SENSE_FORWARD) {
      VERB_FILTER(VERBOSITY_MODERATE, printf("Forward...\n"););

      int p_read_id = 0;
      int read_id;

      VERB_FILTER(VERBOSITY_MODERATE, display_progress(p_read_id, reads_db.size, map->mapped););

#ifdef _OPENMP
#pragma omp parallel for shared(p_read_id, reads_db, ref_index, seeds, quality_mask) schedule(dynamic,256)
#endif
      for (read_id = 0; read_id < reads_db.size; ++read_id) {

#ifdef _OPENMP
        int _ogtn = omp_get_thread_num();
#endif

        /* reseting region filter... */
        reg_filter__reset(FILTER__REF(filter));

        /* set the read */
        ALIGNMENT__RESET_READ(alignment,  reads_db.reads[read_id].sequence, reads_db.reads[read_id].quality);
        /* [NEW] compute read quality mask */
        if (reads_db.reads[read_id].quality) {
          quality__build_mask(reads_db.reads[read_id].quality, reads_db.read_len, QUALITY_MASK_REF);
        }
        /* apply each seed i */
        int si;
        for (si = 0; si < seed_count; ++si) {

          // @filter_efficiency_statistics
#ifdef ENABLE_STATS
          stats__set_seed_idx(si);
#endif

          int read_seed_positions = reads_db.read_len - seeds[si]->length;
          /* on each position j */
          if (seeds[si]->positions) {
            /* if the seed comes with specific positions, use those */
            int j;
            for (j = 0; j < seeds[si]->positions_count; ++j) {
              /* better check if it can be applied, to prevent any overflows */
              if (seeds[si]->positions[j] <= read_seed_positions) {
                process_read_with_seed((reads_db.reads[read_id]).sequence, read_id, ref_id, seeds[si]->positions[j], simd_allowed_diags, FILTER__REF(filter), ref_index[si], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_FORWARD);
              }
            }
          } else {
            /* else, apply it everywhere, filtering according to good quality zones [NEW] */
            int j;
            for (j = 0; j <= read_seed_positions; ++j) {
              if (reads_db.reads[read_id].quality == NULL || quality__accept_seed_relaxed(QUALITY_MASK_REF, reads_db.read_len, seeds[si], j, MIN_TRUSTED_CODES)) {
                process_read_with_seed((reads_db.reads[read_id]).sequence, read_id, ref_id, j, simd_allowed_diags, FILTER__REF(filter), ref_index[si], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_FORWARD);
              }
            }
          }
        }

        /* Do not display this progress bar if other details (such as specific local alignments) are also displayed. */
        VERB_FILTER(VERBOSITY_MODERATE,
                    if ( !(p_read_id & 0x3ff) ) {
                      display_progress(p_read_id, reads_db.size, map->mapped);
                    }
                    );
#ifdef _OPENMP
#pragma omp atomic
#endif
        p_read_id++;

      }

      VERB_FILTER(VERBOSITY_MODERATE, display_progress(reads_db.size, reads_db.size, map->mapped););
      VERB_FILTER(VERBOSITY_NONE, printf("Forward completed in %ld seconds, %d of %d reads aligned.\n\n", time(NULL) - crt_time1, map->mapped, map->size));
    }



    crt_time1 = time(NULL);
    if (ALIGNMENT_SENSE & ALIGNMENT_SENSE_REVERSE) {
      VERB_FILTER(VERBOSITY_MODERATE, printf("Reverse complementary...\n"););

      int p_read_id = 0;
      int read_id;

      VERB_FILTER(VERBOSITY_MODERATE, display_progress(p_read_id, reads_db.size, map->mapped));

#ifdef _OPENMP
#pragma omp parallel for shared(p_read_id, reads_db, ref_index, seeds, quality_mask) schedule(dynamic,256)
#endif
      for (read_id = 0; read_id < reads_db.size; ++read_id) {

#ifdef _OPENMP
        int _ogtn = omp_get_thread_num();
#endif

        /* reseting region filter... */
        reg_filter__reset(FILTER__REF(filter));
        /* set the read */
        read__reverse(&reads_db.reads[read_id], REV_SEQ__REF(tmp_sequence), reads_db.read_len);
        ALIGNMENT__RESET_READ(alignment, REV_SEQ__SEQ_REF(tmp_sequence), REV_SEQ__QUAL_REF(tmp_sequence));
        /* [NEW] compute read quality mask */
        if (reads_db.reads[read_id].quality) {
          quality__build_mask(REV_SEQ__QUAL_REF(tmp_sequence), reads_db.read_len, QUALITY_MASK_REF);
        }
        /* apply each seed i */
        int si;
        for (si = 0; si < seed_count; ++si) {

          // @filter_efficiency_statistics
#ifdef ENABLE_STATS
          stats__set_seed_idx(i);
#endif

          int read_seed_positions = reads_db.read_len - seeds[si]->length;
          /* on each position j */
          if (seeds[si]->positions) {
            /* if the seed comes with specific positions, use those */
            int j;
            for (j = 0; j < seeds[si]->positions_count; ++j) {
              /* better check if it can be applied, to prevent any overflows */
              if (seeds[si]->positions[j] <= read_seed_positions) {
                process_read_with_seed(REV_SEQ__SEQ_REF(tmp_sequence), read_id, ref_id, read_seed_positions - seeds[si]->positions[j], simd_allowed_diags, FILTER__REF(filter), ref_index[si], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_REVERSE);
              }
            }
          } else {
            /* else, apply it everywhere, filtering according to good quality zones [NEW] */
            int j;
            for (j = 0; j <= read_seed_positions; ++j) {
              if (reads_db.reads[read_id].quality == NULL || quality__accept_seed_relaxed(QUALITY_MASK_REF, reads_db.read_len, seeds[si], j, MIN_TRUSTED_CODES)) {
                process_read_with_seed(REV_SEQ__SEQ_REF(tmp_sequence), read_id, ref_id, j, simd_allowed_diags, FILTER__REF(filter), ref_index[si], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_REVERSE);
              }
            }
          }
        }

        /* Do not display this progress bar if other details (such as specific local alignments) are also displayed. */
        VERB_FILTER(VERBOSITY_MODERATE,
                    if ( !(p_read_id & 0x3ff) ) {
                      display_progress(p_read_id, reads_db.size, map->mapped);
                    }
                    );
#ifdef _OPENMP
#pragma omp atomic
#endif
        p_read_id++;

      }

      VERB_FILTER(VERBOSITY_MODERATE, display_progress(reads_db.size, reads_db.size, map->mapped));
      VERB_FILTER(VERBOSITY_NONE, printf("Reverse complementary completed in %ld seconds, %d of %d reads aligned.\n\n", time(NULL) - crt_time1, map->mapped, map->size));
    }

    /* Erase the indexes to make room for the mapping structures  */
    {
      int si;
      for (si = 0; si < seed_count; ++si) {
        index__destroy(ref_index[si]);
      }
    }
  } /* for (ref_id = 0; ref_id < ref_counts; ++ref_id) */


  /* Clear the indexes and the seeds */
  {
    int si;
    for (si = 0; si < seed_count; ++si) {
      // @filter_efficiency_statistics
#ifdef ENABLE_STATS
      stats__display(ref_index[si], FILTER_STATS[si]);
#endif
      free(ref_index[si]);
      ref_index[si] = NULL;
      seed__destroy(seeds[si]);
      free(seeds[si]);
      seeds[si] = NULL;
    }
  }
#ifdef ENABLE_STATS
  stats__destroy();
#endif
  free(ref_index);
  ref_index = NULL;
  free(seeds);
  seeds = NULL;

  ALIGNMENT__DESTROY(alignment);
  REV_SEQ__DESTROY(tmp_sequence);
  QUALITY_MASK__DESTROY;
  FILTER__DESTROY(filter);

  VERB_FILTER(VERBOSITY_NONE, printf("\nAligned %d of %d reads (%5.2f%%) in %ld seconds.\n", map->mapped, map->size, (map->mapped*100.0/map->size), time(NULL) - crt_time));


#ifndef __DONT__MAP__

  /* Generate mapping */
  if (map_unordered == 0) {

    crt_time = time(NULL);

    VERB_FILTER(VERBOSITY_MODERATE, printf("Mapping...\n"););
    GenomeMapType*  genome_map = genome_map__create(map, ref_dbs, ref_dbs_size, &reads_db);
    genome_map__build(genome_map);

    /* It may be useful to see what the alignments look like */
    if (SHOW_TRACEBACK_PATTERNS) {
      display_tracebacks(genome_map);
    }

    /* It may be useful to see which reads didn't make it */
    if (LIST_UNMAPPED_READS) {
      list_unmapped_reads(genome_map);
      list_unmapped_reads_translated(genome_map);
    }

    /* Generate output */
    VERB_FILTER(VERBOSITY_NONE, printf("\nMap built in %ld seconds. Generating output...\n", time(NULL) - crt_time););
    genome_map__generate_SAM_output(genome_map, output);

    /* Erase the genome_map */
    genome_map__destroy(genome_map);
    free(genome_map);
    genome_map = NULL;
  } else {

    /* Generate output */
    VERB_FILTER(VERBOSITY_NONE, printf("\nGenerating output...\n"););
    hit_map__generate_SAM_output(map,
                         &reads_db,
                         ref_dbs, ref_dbs_size,
                         output, map_unordered);
  }
#endif /* __DONT__MAP__ */

  /* Erase the hit_map */
  hit_map__destroy(map);
  free(map);
  map = NULL;


  /* Clear the references [FIXME can done before ? realignment problem on the ref ] */
  {
    int ref_id;
    for (ref_id = 0; ref_id < ref_dbs_size; ++ref_id) {
      ReferenceDBType* ref_db = &ref_dbs[ref_id];
      clear_reference_db(ref_db);
    }
  }

  /* Clear the reads [FIXME can be done before ? quality problem] */
  clear_reads_db(&reads_db);
  free(ref_dbs);
  ref_dbs = NULL;

  VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\nAll done in %ld seconds.\n", time(NULL) - start_time)););
  return RETURN_SUCCESS;
}
